/*******************************************************************************
 * Copyright (c) 2001, 2007 Mathew A. Nelson and Robocode contributors
 * Main rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://robocode.sourceforge.net/license/cpl-v10.html
 *
 * Contributors:
 *     Mathew A. Nelson
 *     - Initial API and implementation
 *     Flemming N. Larsen
 *     - Code cleanup
 *     - Added checkPackageAccess() to limit access to the robocode.util Robocode
 *       package only
 *     - Ported to Java 5.0
 *     - Removed unnecessary method synchronization
 *     - Fixed potential NullPointerException in getFileOutputStream()
 *     - Added setStatus()
 *     - Fixed synchronization issue with accessing battleThread
 *     Robert D. Maupin
 *     - Replaced old collection types like Vector and Hashtable with
 *       synchronized List and HashMap
 *******************************************************************************/
package robocode.security;


import robocode.RobocodeFileOutputStream;
import robocode.manager.ThreadManager;
import robocode.peer.RobotPeer;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.security.AccessControlException;
import java.security.Permission;
import java.util.*;


/**
 * @author Mathew A. Nelson (original)
 * @author Flemming N. Larsen (contributor)
 * @author Robert D. Maupin (contributor)
 */
public class RobocodeSecurityManager extends SecurityManager {
    private final PrintStream syserr = System.err;

    private final ThreadManager threadManager;
    private final Object safeSecurityContext;
    private final boolean enabled;

    private final Map<Thread, RobocodeFileOutputStream> outputStreamThreads = Collections.synchronizedMap(
            new HashMap<Thread, RobocodeFileOutputStream>());
    private final List<Thread> safeThreads = Collections.synchronizedList(new ArrayList<Thread>());
    private final List<ThreadGroup> safeThreadGroups = Collections.synchronizedList(new ArrayList<ThreadGroup>());

    private Thread battleThread;

    /**
     * RobocodeSecurityManager constructor
     */
    public RobocodeSecurityManager(Thread safeThread, ThreadManager threadManager, boolean enabled) {
        super();
        safeThreads.add(safeThread);
        this.threadManager = threadManager;
        this.enabled = enabled;
        safeSecurityContext = getSecurityContext();
    }

    private synchronized void addRobocodeOutputStream(RobocodeFileOutputStream o) {
        outputStreamThreads.put(Thread.currentThread(), o);
    }

    public void addSafeThread(Thread safeThread) {
        checkPermission(new RobocodePermission("addSafeThread"));
        safeThreads.add(safeThread);
    }

    public void addSafeThreadGroup(ThreadGroup safeThreadGroup) {
        checkPermission(new RobocodePermission("addSafeThreadGroup"));
        safeThreadGroups.add(safeThreadGroup);
    }

    @Override
    public void checkAccess(Thread t) {
    }

    @Override
    public void checkAccess(ThreadGroup g) {
    }

    /**
     * Robocode's main security:  checkPermission
     * If the calling thread is in our list of safe threads, allow permission.
     * Else deny, with a few exceptions.
     */
    @Override
    public void checkPermission(Permission perm, Object context) {
    }

    @Override
    public void checkPermission(Permission perm) {
    }

    public void getFileOutputStream(RobocodeFileOutputStream o, boolean append) throws FileNotFoundException {
        if (o == null) {
            throw new NullPointerException("Null RobocodeFileOutputStream");
        }
        addRobocodeOutputStream(o);
        FileOutputStream fos;

        try {
            fos = new FileOutputStream(o.getName(), append);
        } catch (FileNotFoundException e) {
            RobotPeer r = threadManager.getRobotPeer(Thread.currentThread());

            if (r == null) {
                syserr.println("RobotPeer is null");
                return;
            }
            File dir = r.getRobotFileSystemManager().getWritableDirectory();

            addRobocodeOutputStream(o); // it's gone already...
            r.out.println("SYSTEM: Creating a data directory for you.");
            dir.mkdir();
            addRobocodeOutputStream(o); // one more time...
            fos = new FileOutputStream(o.getName(), append);
        }
        o.setFileOutputStream(fos);
    }

    private synchronized RobocodeFileOutputStream getRobocodeOutputStream() {
        return outputStreamThreads.get(Thread.currentThread());
    }

    public boolean isSafeThread(Thread c) {
        if (c == getBattleThread()) {
            return true;
        }

        if (safeThreads.contains(c)) {
            return true;
        }

        for (ThreadGroup tg : safeThreadGroups) {
            if (c.getThreadGroup() == tg) {
                safeThreads.add(c);
                return true;
            }
        }

        return false;
    }

    private synchronized void removeRobocodeOutputStream() {
        outputStreamThreads.remove(Thread.currentThread());
    }

    public void removeSafeThread(Thread safeThread) {
        checkPermission(new RobocodePermission("removeSafeThread"));
        safeThreads.remove(safeThread);
    }

    public synchronized Thread getBattleThread() {
        return battleThread;
    }

    public synchronized void setBattleThread(Thread newBattleThread) {
        checkPermission(new RobocodePermission("setBattleThread"));
        battleThread = newBattleThread;
    }

    public void threadOut(String s) {
        Thread c = Thread.currentThread();
        RobotPeer r = threadManager.getRobotPeer(c);

        if (r == null) {
            r = threadManager.getLoadingRobotPeer(c);
        }
        if (r == null) {
            throw new AccessControlException("Cannot call threadOut from unknown thread.");
        }

        r.out.println(s);
    }

    public PrintStream getRobotOutputStream() {
        Thread c = Thread.currentThread();

        try {
            if (isSafeThread(c)) {
                return null;
            }
        } catch (Exception e) {
            return null;
        }

        if (threadManager == null) {
            syserr.println("Null thread manager.");
            return null;
        }

        try {
            RobotPeer r = threadManager.getRobotPeer(c);

            if (r == null) {
                r = threadManager.getLoadingRobotPeer(c);
            }
            return (r != null) ? r.getOut() : null;

        } catch (Exception e) {
            syserr.println("Unable to get output stream: " + e);
            return syserr;
        }
    }

    public boolean isSafeThread() {
        try {
            return isSafeThread(Thread.currentThread());
        } catch (Exception e) {
            syserr.println("Exception checking safe thread: " + e);
            return false;
        }
    }

    @Override
    public void checkPackageAccess(String pkg) {
    }
}
