package ru.ifmo.robocode.gep;

import ru.ifmo.robocode.AutomataFunction;
import ru.ifmo.robocode.Environment;
import ru.ifmo.robocode.Basis;
import ru.ifmo.robocode.ga.*;

public class GepConfiguration extends Configuration<Chromosome> {

    public static final int NUM_EXPR = AutomataFunction.NUM;
    public static final int MAX_ARITY = 4;

    public static final int NUM_FUNCTIONS = Basis.values().length;
    public static final int NUM_TERMS = Environment.terms().length;

    private static final int HEAD_LEN = 10;
    private static final int TAIL_LEN = HEAD_LEN * (MAX_ARITY - 1) + 1;
    public static final int LEN = TAIL_LEN + HEAD_LEN;

    private static final double HEAD_FUNC_RATE = 0.95; // prob for func symb in head

    protected Crossover<Chromosome> createCrossover() {
        return new Crossover<Chromosome>() {
            private final static double RATE = 0.6;
            private final static double RATE_SINGLE = 0.1;
            private final static double RATE_INVERTION = 0.1;

            private void crossover(IntString p1, IntString p2, IntString c) {
                boolean[] first = new boolean[p1.length()];
                first[0] = true;
                double inv = RATE_INVERTION / first.length;
                for (int i = 1; i < first.length; i++) {
                    first[i] = first[i - 1];
                    if (Math.random() < inv) {
                        first[i] = !first[i];
                    }
                }
                for (int i = 0; i < c.length(); i++) {
                    c.set(i, first[i] ? p1.get(i) : p2.get(i));
                }
            }

            public void crossover(Chromosome p1, Chromosome p2, Chromosome c) {
                if (Math.random() > RATE) {
                    c.copy(Math.random() > 0.5 ? p1 : p2);
                    return;
                }

                for (int i = 0; i < p1.f.length; i++) {
                    if (Math.random() < RATE_SINGLE) {
                        crossover(p1.f[i], p2.f[i], c.f[i]);
                    } else {
                        c.f[i].copy(Math.random() > 0.5 ? p1.f[i] : p2.f[i]);
                    }
                }
            }
        };
    }

    protected Mutation<Chromosome> createMutation() {
        return new Mutation<Chromosome>() {
            private final static double RATE = 0.1;   // percentage of bits to change

            public void mutate(Chromosome c) {
                for (IntString aF : c.f) {
                    mutate(aF);
                }
            }

            private void mutate(IntString c) {
                final double prob = RATE / c.length();
                for (int i = 0; i < HEAD_LEN; i++) {
                    if (Math.random() < prob) {
                        c.set(i, Math.random() < HEAD_FUNC_RATE ? rndFunc() : rndTerm());
                    }
                }
                for (int i = HEAD_LEN; i < LEN; i++) {
                    if (Math.random() < prob) {
                        c.set(i, rndTerm());
                    }
                }
            }

        };
    }

    protected Randomizer<Chromosome> createRandomizer() {
        return new Randomizer<Chromosome>() {

            private IntString randomizeIntString(int length) {
                IntString c = new IntString(length);
                for (int i = 0; i < HEAD_LEN; i++) {
                    c.set(i, Math.random() < HEAD_FUNC_RATE ? rndFunc() : rndTerm());
                }
                for (int i = HEAD_LEN; i < LEN; i++) {
                    c.set(i, rndTerm());
                }
                return c;
            }

            public Chromosome randomize(int length) {
                IntString s[] = new IntString[NUM_EXPR];
                for (int i = 0; i < s.length; i++) {
                    s[i] = randomizeIntString(length);
                }
                return new Chromosome(s);
            }

        };
    }

    protected Selector createSelector() {
        return SelectorFactory.getInstance().rouletteSelector();
    }

    protected Utils<Chromosome> createUtils() {
        return new Utils<Chromosome>() {
            private double diff(IntString c1, IntString c2) {
                double r = 0;
                for (int i = 0; i < c1.length(); i++) {
                    if (c1.get(i) != c2.get(i)) {
                        r++;
                    }
                }
                return r / c1.length();
            }

            public void copy(Chromosome src, Chromosome dst) {
                dst.copy(src);
            }

            public double diff(Chromosome c1, Chromosome c2) {
                double d = 0;
                for (int i = 0; i < NUM_EXPR; i++) {
                    d += diff(c1.f[i], c2.f[i]);
                }
                return d / NUM_EXPR;
            }

            public Chromosome allocate(int length) {
                IntString s[] = new IntString[NUM_EXPR];
                for (int i = 0; i < NUM_EXPR; i++) {
                    s[i] = new IntString(length);
                }
                return new Chromosome(s);
            }
        };
    }

    private int rndTerm() {
        return NUM_FUNCTIONS + (int) (Math.random() * NUM_TERMS);
    }

    private int rndFunc() {
        return (int) (Math.random() * NUM_FUNCTIONS);
    }

}
